/* sdsl - succinct data structures library
    Copyright (C) 20010 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file algorithms_for_compressed_suffix_arrays.hpp
    \brief algorithms_for_compressed_suffix_arrays.hpp contains algorithms for compressed suffix arrays.
	\author Simon Gog
*/ 
#ifndef INCLUDED_SDSL_ALGORITHMS_FOR_COMPRESSED_SUFFIX_ARRAYS
#define INCLUDED_SDSL_ALGORITHMS_FOR_COMPRESSED_SUFFIX_ARRAYS

#include "int_vector.hpp" // for bit_vector
#include <stack> // for calculate_supercartesian_tree_bp

namespace sdsl{

namespace algorithm{	

template<class Csa>	
void set_text(const unsigned char *str, typename Csa::size_type len, typename Csa::size_type *C, unsigned char *char2comp, unsigned char *comp2char, uint8_t &sigma){
	typedef typename Csa::size_type size_type;
	for(uint16_t i=0; i<256; ++i)
		char2comp[i] = comp2char[i] = C[i] = 0;
	C[256] = 0;
	if( str == NULL or len ==0 )
		return;
	const unsigned char *p = str;
	for(size_type i=0;i<len-1;++i){
		char2comp[*p++] = 1;
	}
	assert(char2comp[0] == 0);
	char2comp[*p] = 1;
	assert(char2comp[0] == 1);
	size_type value = 0;
	for(uint16_t i=0; i<256; ++i)
		if(char2comp[i])
			char2comp[i] = value++;
	sigma = value;
	for(uint16_t i=0; i<256; ++i)
		if(char2comp[i])
			comp2char[char2comp[i]] = i;
	for(uint16_t i=0; i<257; ++i) C[i]=0;
	for(size_type i=0; i<len; ++i) ++C[char2comp[str[i]]];
	for(uint16_t i=256;i>0; --i) C[i] = C[i-1];
	C[0] = 0;
	for(uint16_t i=1; i<257; ++i) C[i] += C[i-1];	
	if(C[256]!=len){
		std::cerr<<"C[256]="<<C[256]<<" "<<len<<std::endl;
	}
	assert(C[256]==len);
	assert(C[sigma+1]==len);

}

//! Calculates if a character c occures in the text of the compressed suffix array c.
/*!
 * \param csa 	The csa in which we search for the occurence of c.
 * \param c		The character c for which we search in the text of the compressed suffix array.
 * \return 		True, if c occures in the text of the compressed suffix array and False otherwise.
 * \pre \f$ csa.size()>0 \f$ 
 */ 
template<class Csa>
bool char_occures_in_text_of_csa(const Csa &csa, typename Csa::char_type c){
	return (csa.char2comp[c] > 0) or (csa.char2comp[c]==c); 
}


//template<class Csa>
//bool char_at_char_pos_equals_char(const Csa &csa, typename Csa::size_type char_pos, typename Csa::char_type c){
//	typename Csa::
//}

/*
 * \par Time complexity
 *    \f$ \Order{\log \sigma} \f$
*/ 
template<class Csa>	
const unsigned char get_ith_character_of_the_first_row(const typename Csa::size_type i, const Csa &csa){
		assert( i < csa.size() );
		if( csa.sigma < 16 ){ //<- if sigma is small search linear
			typename Csa::size_type res=1;
			while( csa.C[res] <= i )
				++res;
			return csa.comp2char[res-1];
		}
		else
		{	// binary search the character with C 
			typename Csa::size_type upper_c = csa.sigma, lower_c = 0; // lower_c inclusive, upper_c exclusive
			typename Csa::size_type res=0;
			do{
				res = (upper_c+lower_c)/2;
				if( i < csa.C[res] ){
					upper_c = res;	
				}else if( i >= csa.C[res+1] ){
					lower_c = res+1;
				}
			}
			while(  i < csa.C[res] or i >= csa.C[res+1]  ); // i is not in the interval
			return csa.comp2char[res];
		}
}

template<class WaveletTree>
typename WaveletTree::size_type psi_from_wavelet_tree(const WaveletTree &wt, typename WaveletTree::size_type i){
	assert( i < wt.size() );
	return 0;	
}

}// end namespace algorithm

}// end namespace sdsl

#endif

