/* sdsl - succinct data structures library
    Copyright (C) 2009 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file algorithms_for_compressed_suffix_trees.hpp
    \brief algorithms_for_compressed_suffix_trees.hpp contains algorithms for compressed suffix trees.
	\author Simon Gog
*/ 
#ifndef INCLUDED_SDSL_ALGORITHMS_FOR_COMPRESSED_SUFFIX_TREES
#define INCLUDED_SDSL_ALGORITHMS_FOR_COMPRESSED_SUFFIX_TREES

#include "int_vector.hpp" // for bit_vector
#include "sorted_stack_support.hpp" // for construct_supercartesian_tree_bp
#include <stack> // for calculate_supercartesian_tree_bp


namespace sdsl{

namespace algorithm{	

//! Calculate the balanced parentheses of the Super-Cartesian tree, described in Ohlebusch and Gog (SPIRE 2009). 
/*! \param vec Random access container for which the Super-Cartesian tree representation should be calculated. 
 *             The value_type of vec should be an unsigned integer type.
 *  \param bp Reference to the balanced parentheses sequence which represents the Super-Cartesian tree.
 *  \param minimum Specifies if the higher levels contains minima or maxima. Default is maxima.
 *  \par Time complexity
 *       \f$ \Order{2n} \f$, where \f$ n=\f$vec.size()  
 *  \par Space complexity
 *       \f$ \Order{n \cdot \log n } \f$ bits.
 */	
template<class RandomAccessContainer>	
void construct_supercartesian_tree_bp(const RandomAccessContainer &vec, bit_vector &bp, const bool minimum=true){
	typedef typename RandomAccessContainer::size_type size_type;
	bp.resize( 2*vec.size() );    // resize bit vector for balanaced parantheses to 2 n bits
	std::stack<typename RandomAccessContainer::value_type> vec_stack;

	size_type k=0;
	for(size_type i=0; i < vec.size(); ++i){
		typename RandomAccessContainer::value_type l = vec[i];
		if( minimum ){
			while( vec_stack.size() > 0 and l < vec_stack.top() ){
				vec_stack.pop(); bp[k++] = 0; // writing a closing parenthesis
			}
		}else{
			while( vec_stack.size() > 0 and l > vec_stack.top() ){
				vec_stack.pop(); bp[k++] = 0; // writing a closing parenthesis
			}	
		}
		vec_stack.push(l);
		bp[k++] = 1; // writing an opening  parenthesis
	}
	while( vec_stack.size() > 0 ){
		vec_stack.pop();
		bp[k++] = 0; // writing a closing parenthesis
	}
	assert( k == 2*vec.size() );
}

//! Calculate the balanced parentheses of the Super-Cartesian tree, described in Ohlebusch and Gog (SPIRE 2009). 
/*! \param vec Random access container for which the Super-Cartesian tree representation should be calculated. 
 *             The value_type of vec should be an unsigned integer type.
 *  \param bp Reference to the balanced parentheses sequence which represents the Super-Cartesian tree.
 *  \param minimum Specifies if the higher levels contains minima or maxima. Default is maxima.
 *  \par Time complexity
 *       \f$ \Order{2n} \f$, where \f$ n=\f$vec.size()  
 *  \par Space complexity
 *       \f$\Order{n}\f$ bits, by the stack_support described in the paper "Optimal Succinctness For Range Minimum Queries" of Johannes Fischer.
 */	
template<class RandomAccessContainer>	
void construct_supercartesian_tree_bp_succinct(const RandomAccessContainer &vec, bit_vector &bp, const bool minimum=true){
	typedef typename RandomAccessContainer::size_type size_type;
	bp.resize( 2*vec.size() );    // resize bit vector for balanaced parantheses to 2 n bits
	sorted_stack_support<RandomAccessContainer> vec_stack(&vec);

	size_type k=0;
	for(size_type i=0; i < vec.size(); ++i){
		if( minimum ){
			while( vec_stack.size() > 0 and vec[i] < vec[vec_stack.top()] ){
				vec_stack.pop(); bp[k++] = 0; // writing a closing parenthesis
			}
		}else{
			while( vec_stack.size() > 0 and vec[i] > vec[vec_stack.top()] ){
				vec_stack.pop(); bp[k++] = 0; // writing a closing parenthesis
			}	
		}
		vec_stack.push(i);
		bp[k++] = 1; // writing an opening  parenthesis
	}
	while( vec_stack.size() > 0 ){
		vec_stack.pop();
		bp[k++] = 0; // writing a closing parenthesis
	}
	assert( k == 2*vec.size() );
}

template<class RandomAccessContainer>
void construct_first_p_index(const RandomAccessContainer &vec, bit_vector &bp, const bool minimum=true){
	typedef typename RandomAccessContainer::size_type size_type;
	bp = bit_vector(vec.size(), 0);
//	std::cerr<<"bp.size()="<<bp.size()<<std::endl;
	sorted_stack_support<RandomAccessContainer> vec_stack(&vec);
	size_type k=0;
	for(size_type i=0, t; i < vec.size(); ++i){
//		std::cerr<<"k="<<k<<std::endl;
		if( minimum ){
//			std::cerr<<"vec_stack.size()="<<vec_stack.size()<<std::endl;
//			if(!vec_stack.empty())
//				std::cerr<<"vec_stack.top()="<<vec_stack.top()<<std::endl;
//			else
//				std::cerr<<"stack empty"<<std::endl;
			while( vec_stack.size() > 0 and vec[i] < vec[vec_stack.top()] ){
				t = vec[vec_stack.top()];
				vec_stack.pop();
//				std::cerr<<"t="<<t<<" k="<<k<<std::endl;
				if( vec_stack.size() == 0 or t != vec[vec_stack.top()] )
					bp[k] = 1;
				++k;

			}
//			std::cerr<<"done"<<std::endl;
		}else{
			while( vec_stack.size() > 0 and vec[i] > vec[vec_stack.top()] ){
				t = vec[vec_stack.top()];
				vec_stack.pop(); 
				if( vec_stack.size() == 0 or t != vec[vec_stack.top()] )
					bp[k] = 1;
				++k;
			}	
		}
//		std::cerr<<"i="<<i<<" vec[i]="<<vec[i]<<std::endl;
		vec_stack.push(i);
	}
	while( vec_stack.size() > 0 ){
		size_type t = vec[vec_stack.top()];
		vec_stack.pop();
		if( vec_stack.size() == 0 or t != vec[vec_stack.top()] )
			bp[k] = 1;
		++k;
	}
	assert( k == vec.size() );
}

}// end namespace algorithm

}// end namespace sdsl

#endif

