/* sdsl - succinct data structures library
    Copyright (C) 2009 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file cst_iterators.hpp
    \brief cst_iterators.hpp contains iterator classes for traversing (compressed) suffix arrays.
	\author Simon Gog
*/
#ifndef INCLUDED_SDSL_CST_ITERATORS
#define INCLUDED_SDSL_CST_ITERATORS

namespace sdsl{

//! An forward iterator for (compressed) suffix trees.
/*! The cst_dfs_const_forward_iterator iterates through the nodes of a (compressed) suffix tree in 
    depth first search (dfs) order. Note, that 
	  - each inner node is visited twice, and
	  - each leaf node is visited only once.

	If the current node is a inner node, the method visit() returns 1 for the first visit and 2 for
    the second one.	

	\par Time complexity
		\f$\Order{1}\f$ for all methods 
	\par Space complexity
		\f$\Order{1} \f$	

	This iterator is the standard iterator for the classes
		- sdsl::cst_sada, and
		- sdsl::cst_sct	
 */
template<class Cst>
class cst_dfs_const_forward_iterator: public std::iterator<std::forward_iterator_tag, typename Cst::node_type>{
	public:
	typedef typename Cst::node_type value_type;
	typedef const value_type const_reference;
	typedef typename Cst::size_type size_type;
	typedef cst_dfs_const_forward_iterator<Cst> iterator;
	private:
	const Cst *m_cst; 
	typename Cst::node_type m_v;
	bool	 m_visited;	
	bool 	 m_valid;
	public:

	//! Default constructor	
	cst_dfs_const_forward_iterator():m_cst(NULL),m_visited(false),m_valid(false){}

	//! Constructor
	cst_dfs_const_forward_iterator(const Cst *cst, const value_type node, bool visited=false, bool valid=true):m_visited(visited), m_valid(valid){
		m_cst = cst;
		m_v = node;
		if( m_cst == NULL )
			m_valid = false;
	}

	//! Returns how often the current node was already visited.
	uint8_t visit(){
		return 1+(uint8_t)m_visited;
	}

	//! Method for dereferencing the iterator.
	const_reference operator*()const{
		return m_v;
	}

	//! Prefix increment of the iterator.
	iterator& operator++(){
		if(!m_valid)
			return *this;
		if( m_v == m_cst->root() and m_visited ){
			m_valid = false;
			return *this;
		}
		value_type w;
		if( !m_visited ){
			if( m_cst->is_leaf(m_v) ){
				w = m_cst->sibling(m_v);
				if( w == m_cst->root() ){
					w = m_cst->parent(m_v);
					m_visited = true;
				}
			}else{
				w = m_cst->ith_child(m_v, 1);
			}
		}else{
			w = m_cst->sibling(m_v);
			if( w == m_cst->root() ){
				w = m_cst->parent(m_v);
			}else{
				m_visited = false;
			}
		}
		m_v = w;
		return *this;
	}

	//! Postfix increment of the iterator.
	iterator operator++(int x){
		iterator it = *this;
		++(*this);
		return it;
	}

	//! Equality operator.
	bool operator==(const iterator &it)const{
		return it.m_cst == m_cst && it.m_visited == m_visited && it.m_valid == m_valid;
	}

	//! Inequality operator.
	bool operator!=(const iterator &it)const{
		return !(*this==it);
	}

};

} // end namespace sdsl

#endif
