/* sdsl - succinct data structures library
    Copyright (C) 2009 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file lcp_sada.hpp
    \brief lcp_sada.hpp contains an implementation of a compressed lcp array. 
	\author Simon Gog
*/
/* Changes:
    - Removed unnecessary rank support
 */
#ifndef INCLUDED_SDSL_LCP_SADA
#define INCLUDED_SDSL_LCP_SADA

#include "lcp.hpp"
#include "int_vector.hpp" 
#include "algorithms.hpp"
#include "iterators.hpp"
#include "csa_sada_prac.hpp"  // for standard template initialization of lcp_sada 
#include "rank_support.hpp"   // for standard template initialization of lcp_sada 
#include "select_support.hpp" // for standard template initialization of lcp_sada
#include <iostream>
#include <algorithm>
#include <cassert>
#include <cstring> // for strlen
#include <iomanip>
#include <iterator>

namespace sdsl{

//! A class for the compressed version of lcp information of an suffix array of class Csa proposed by Sadakane in the paper "Succinct Representation of lcp Information and Imporvements in the Compressed Suffix Arrays".
/*! The class has three template parameters: 
     - \a Csa is the (compressed) suffix array for which the lcp information should by provided. 
     - \a SelectSupport is the SelectSupport class which is used in the data structure to calculate the select queries.
	The representation of the lcp information corresponds to the concept of an imutable random access container of the STL. 
 */
template<class Csa = csa_sada_prac<>, class SelectSupport = select_support_mcl<1> >
class lcp_sada{
	public:
	typedef typename Csa::value_type					 value_type;	// STL Container requirement
	typedef random_access_const_iterator<lcp_sada>		 const_iterator;// STL Container requirement
	typedef const_iterator 								 iterator;		// STL Container requirement
	typedef const value_type							 const_reference;
	typedef const_reference								 reference;
	typedef const_reference*							 pointer;
	typedef const pointer								 const_pointer;
	typedef size_t 										 size_type;		// STL Container requirement
	typedef ptrdiff_t  									 difference_type; // STL Container requirement
	typedef Csa											 csa_type;
	typedef const unsigned char*						 pattern_type;
	private:
		// 
		const Csa 		*m_csa;
		bit_vector      m_data;
		SelectSupport  	m_select_support;
	public:
		const Csa *&csa;
		//! Default Constructor
		lcp_sada(): csa(m_csa){}
		//! Default Destructor
		~lcp_sada(){}
		//! Copy constructor
		lcp_sada(const lcp_sada<Csa, SelectSupport> &lcp_c):csa(m_csa)
		{ 
			m_csa = lcp_c.m_csa;
			m_data = lcp_c.m_data;
			m_select_support = lcp_c.m_select_support;
			m_select_support.set_vector(&m_data);
		}

		//! Constructor for the compressed lcp from a compressed suffix array.
		template<class Text, class Sa>
		lcp_sada(const Csa *csa, const Text &text, const Sa &sa);

		void set_csa(const Csa *f_csa){
			m_csa = f_csa;
		}

//		const EncVector & psi() const;

		//! Number of elements in the instance.
		/*! Required for the Container Concept of the STL.
		 *  \sa max_size, empty
		 */
		size_type size()const{ return m_csa->size();}

		//! Returns the largest size that lcp_sada can ever have.
		/*! Required for the Container Concept of the STL.
		 *  \sa size
		 */
		static size_type max_size(){ return Csa::max_size(); }

		//! Returns if the data strucutre is empty.
		/*! Required for the Container Concept of the STL.A
		 * \sa size
		 */
		bool empty()const{ return m_csa->empty();}

		//! Swap method for lcp_sada
		/*! The swap method can be defined in terms of assignment.
			This requires three assignments, each of which, for a container type, is linear
			in the container's size. In a sense, then, a.swap(b) is redundant.
			This implementation guaranties a run-time complexity that is constant rather than linear.
			\param lcp_c lcp_sada to swap.

			Required for the Assignable Conecpt of the STL.
		  */
		void swap(lcp_sada<Csa, SelectSupport> &lcp_c);

		//! Returns a const_iterator to the first element.
		/*! Required for the STL Container Concept.
		 *  \sa end
		 */
		const_iterator begin()const;

		//! Returns a const_iterator to the element after the last element.
		/*! Required for the STL Container Concept.
		 *  \sa begin.
		 */
		const_iterator end()const;

		//! []-operator
		/*! \param i Index of the value. \f$ i \in [0..size()-1]\f$.
		 * Time complexity: O(suffix array access)
		 * Required for the STL Random Access Container Concept.
		 */
		inline value_type operator[](size_type i)const;

		//! Assignment Operator.
		/*!
		 *	Required for the Assignable Concept of the STL.
		 */
		lcp_sada& operator=(const lcp_sada &lcp_c);

		//! Equality Operator
		/*! Two Instances of lcp_sada are equal if
		 *  all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator!=
		 */
		bool operator==(const lcp_sada &lcp_c)const;

		//! Unequality Operator
		/*! Two Instances of lcp_sada are equal if
		 *  not all their members are equal.
		 *  \par Required for the Equality Comparable Concept of the STL.
		 *  \sa operator==
		 */
		bool operator!=(const lcp_sada &lcp_c)const;

		//! Serialize to a stream.
		/*! \param out Outstream to write the data structure.
		 *  \return The number of written bytes.
		 */
		size_type serialize(std::ostream &out) const;

		//! Load from a stream.
		/*! \param in Inputstream to load the data structure from.
		 *  \param csa Compressed Suffix Array that is supported.
		 */
		void load(std::istream &in, const Csa *csa);
};

// == template functions ==

template<class Csa, class SelectSupport>
template<class Text, class Sa>
lcp_sada<Csa, SelectSupport>::lcp_sada(const Csa *csa, const Text &text, const Sa &sa):csa(m_csa){
	m_csa = csa;
	if( m_csa == NULL or m_csa->size() == 0 ){
		return;
	}
	assert(csa->size() == sa.size());
#ifdef SDSL_DEBUG
	std::cerr<<"start building lcp_sada"<<std::endl;
#endif	
	m_data = bit_vector(2*sa.size(), 0);
	typename Csa::size_type sa_1 = sa(0), l=0, oldl = 0;	
	for(typename Csa::size_type i=0,j=0,data_cnt=0; i < sa.size();++i){
		oldl = l;
		if(l) --l;
		if( sa_1 ){
			j = sa[sa_1-1];	
			while(i+l < sa.size() and j+l < sa.size() and text[i+l]==text[j+l]) ++l;
		}
		else{
			l = 0;
		}
		if(i>0)
			data_cnt += l-oldl+1;
		else
			data_cnt = l;
		m_data[data_cnt]=1;
		++data_cnt;
		sa_1 = sa.psi[sa_1];
	}
	m_select_support.init(&m_data);
#ifdef SDSL_DEBUG
	std::cerr<<"finished building lcp_sada"<<std::endl;
#endif	
}

template<class Csa, class SelectSupport>
void lcp_sada<Csa, SelectSupport>::swap(lcp_sada<Csa, SelectSupport> &lcp_c){
	std::swap(m_csa, lcp_c.m_csa);   // TODO: check if this is necessary
	m_data.swap(lcp_c.m_data);
	m_select_support.swap(lcp_c.m_select_support); // select swap after the swap of the bit vector m_data
}

template<class Csa, class SelectSupport>
inline typename lcp_sada<Csa, SelectSupport>::value_type lcp_sada<Csa, SelectSupport>::operator[](size_type i)const{
	size_type j = (*m_csa)[i];
	size_type s = m_select_support.select( j+1 );
	return s-(j<<1)/*-m_rank_support.rank(s)*/;
}


template<class Csa, class SelectSupport>
typename lcp_sada<Csa, SelectSupport>::size_type lcp_sada<Csa, SelectSupport>::serialize(std::ostream &out) const{
	size_type written_bytes = 0;
	written_bytes += m_data.serialize(out);
	written_bytes += m_select_support.serialize(out);
	return written_bytes;
}

template<class Csa, class SelectSupport>
void lcp_sada<Csa, SelectSupport>::load(std::istream &in, const Csa *csa) {
	m_csa = csa;
	m_data.load(in);
	m_select_support. load(in, &m_data);
}


template<class Csa, class SelectSupport>
lcp_sada<Csa, SelectSupport>& lcp_sada<Csa, SelectSupport>::operator=(const lcp_sada &lcp_c) {
	if(this != &lcp_c ){
		m_csa 			= lcp_c.m_csa;
		m_data 			= lcp_c.m_data;
		m_select_support= lcp_c.m_select_support;
		m_select_support.set_vector(&m_data);
	}
	return *this;	
}


template<class Csa, class SelectSupport>
bool lcp_sada<Csa, SelectSupport>::operator==(const lcp_sada &lcp_c)const{
	if( this == &lcp_c )
		return true;
	return m_csa == lcp_c.m_csa and m_data == lcp_c.m_data and m_select_support == lcp_c.m_select_support;
}

template<class Csa, class SelectSupport>
bool lcp_sada<Csa, SelectSupport>::operator!=(const lcp_sada &lcp_c)const{
	return !(*this == lcp_c);
}

template<class Csa, class SelectSupport>
typename lcp_sada<Csa, SelectSupport>::const_iterator lcp_sada<Csa, SelectSupport>::begin()const{
	return const_iterator(this, 0);
}

template<class Csa, class SelectSupport>
typename lcp_sada<Csa, SelectSupport>::const_iterator lcp_sada<Csa, SelectSupport>::end()const{
	return const_iterator(this, size());
}



} // end namespace sdsl

#endif
