/* sdsl - succinct data structures library
    Copyright (C) 2009 Simon Gog 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/ .
*/
/*! \file wavelet_tree.hpp
    \brief wavelet_tree.hpp contains a generic class for the wavelet tree proposed first by Grossi et al. 2003 and applied to the BWT in Foschini et al. 2004.
	\author Simon Gog
*/
#ifndef INCLUDED_SDSL_WAVELET_TREE
#define INCLUDED_SDSL_WAVELET_TREE

#include "int_vector.hpp"
#include "rank_support_v.hpp"
#include "select_support_mcl.hpp"
#include "bitmagic.hpp"
#include <set> // for calculating the alphabet size
#include <map> // for mapping a symbol to its lexicographical index
#include <algorithm> // for std::swap
#include <stdexcept>

#ifdef SDSL_DEBUG
	#define SDSL_DEBUG_WAVELET_TREE
#endif

//! Namespace for the succinct data structure library.
namespace sdsl{

struct unsigned_char_map{
	unsigned char m_map[256];

	unsigned char& operator[](unsigned char i){
		return *(m_map+i);
	}

	unsigned char operator[](unsigned char i)const{
		return m_map[i];
	}

	void clear(){
		for(uint16_t i=0; i<256; ++i)
			m_map[i] = 0;
	}

	uint16_t serialize(std::ostream &out)const{
		uint16_t written_bytes = 0;
		for(uint16_t i=0; i<256; ++i){
			out.write((char*)&m_map[i], sizeof(m_map[i]));
			written_bytes += sizeof(m_map[256]);
		}
		return written_bytes;
	}

	void load(std::istream &in){
		for(uint16_t i=0; i<256; ++i){
			in.read((char*) &m_map[i], sizeof(m_map[i]));
		}
	}

	void swap(unsigned_char_map &map){
		if( this != &map ){
			for(uint16_t i=0; i<256; ++i){
				std::swap(m_map[i], map.m_map[i]);
			}		
		}
	}
};	

template<class RandomAccessContainer>	
class wavelet_tree_trait{
	public:
		typedef typename RandomAccessContainer::size_type 		size_type;	
		typedef typename RandomAccessContainer::value_type 		value_type;	
		typedef RandomAccessContainer& 							reference_type;
		typedef std::map<value_type, size_type>					map_type;
		typedef std::map<value_type, size_type>					inv_map_type;

		static size_type alphabet_size_and_map(const reference_type rac, size_type n, map_type &map, inv_map_type &inv_map, value_type &first_symbol){
			if( n > 0 )
				first_symbol = rac[0];
			map.clear();
			inv_map.clear();
			size_type alphabet_size = 0;
			for(size_type i=0; i<n; ++i){
				if( map.find(rac[i]) == map.end() ){
					map[rac[i]]	= alphabet_size;
					inv_map[alphabet_size] = rac[i];
					++alphabet_size;
				}
			}
			return alphabet_size;
		}

		static bool symbol_available(const map_type &map, const value_type c, const value_type first_symbol){
			return map.find(c)!=map.end();
		}

		static size_type serialize_maps(std::ostream &out, const map_type &map, const inv_map_type &inv_map){
			throw std::logic_error( util::demangle(typeid(wavelet_tree_trait<RandomAccessContainer>).name())+": serialize not implemented" );
			return 0;	
		}

		static size_type load_maps(std::istream &in, map_type &map, inv_map_type &inv_map){
			throw std::logic_error( util::demangle(typeid(wavelet_tree_trait<RandomAccessContainer>).name())+": load not implemented" );
			return 0;
		}
};

template<class character>
class wavelet_tree_trait<character*>{
	public:
		typedef bit_vector::size_type 			size_type;  
		typedef character						value_type;
		typedef character*						reference_type;
		typedef std::map<value_type, size_type>	map_type;
		typedef std::map<value_type, size_type>	inv_map_type;

		static size_type alphabet_size_and_map(const reference_type rac, size_type n, map_type &map, inv_map_type &inv_map, value_type &first_symbol){
			if( n > 0 )
				first_symbol = *(rac+0);
			map.clear();
			inv_map.clear();
			size_type alphabet_size = 0;
			for(size_type i=0; i<n; ++i){
				if( map.find(*(rac+i)) == map.end() ){
					map[*(rac+i)]	= alphabet_size;
					inv_map[alphabet_size] = *(rac+i);
					++alphabet_size;
				}
			}
			return alphabet_size;
		}

		static bool symbol_available(const map_type &map, const value_type c, const value_type first_symbol){
			return map.find(c)!=map.end();
		}

		static size_type serialize_maps(std::ostream &out, const map_type &map, const inv_map_type &inv_map){
			throw std::logic_error( util::demangle(typeid(wavelet_tree_trait<character*>).name())+": serialize not implemented" );
			return 0;
		}

		static size_type load_maps(std::istream &in, map_type &map, inv_map_type &inv_map){
			throw std::logic_error( util::demangle(typeid(wavelet_tree_trait<character*>).name())+": load not implemented" );
			return 0;
		}
};

template<>
class wavelet_tree_trait<unsigned char*>{
	public:
		typedef bit_vector::size_type 		size_type;  
		typedef unsigned char				value_type;
		typedef unsigned char*				reference_type;
		typedef unsigned_char_map			map_type;
		typedef unsigned_char_map			inv_map_type;

		static size_type alphabet_size_and_map(const reference_type rac, size_type n, map_type &map, inv_map_type &inv_map, value_type &first_symbol){
			map.clear();
			inv_map.clear();
			if(n==0)
				return 0;
			first_symbol	= *rac;
			map[*rac] = 0; 
			inv_map[0] = *rac;
			size_type alphabet_size = 1;
			for(size_type i=1; i<n; ++i) {
				value_type c = *(rac+i);
				if( c != first_symbol and map[c]==0 ){
					map[c] = alphabet_size;
					inv_map[alphabet_size] = c;
					++alphabet_size;
				}
			}
			return alphabet_size;		
		}

		static bool symbol_available(const map_type &map, const value_type c, const value_type first_symbol){
			return map[c] > 0 or (c==first_symbol);
		}

		static size_type serialize_maps(std::ostream &out, const map_type &map, const inv_map_type &inv_map){
			size_type written_bytes = 0;
			written_bytes += map.serialize(out);
			written_bytes += inv_map.serialize(out);
			return written_bytes;
		}

		static void load_maps(std::istream &in, map_type &map, inv_map_type &inv_map){
			map.load(in);
			inv_map.load(in);
		}
};

//! A wavelet tree class 
/*!
 * A wavelet tree is build for a vector of characters over the alphabet \f$\Sigma\f$.
 * The wavelet tree \f$wt\f$ consits of a tree of bit vector and provides three efficient methods:
 *   - The "[]"-operator: \f$wt[i]\f$ returns the ith symbol of vector for which the wavelet tree was build for.
 *   - The rank method: \f$wt.rank(i,c)\f$ returns the number of occurences of symbol \f$c\f$ in the prefix [0..i-1] in the vector for which the wavelet tree was build for.
 *   - The select method: \f$wt.select(j,c)\f$ returns the index \f$i\in [0..size()-1]\f$ of the jth occurence of symbol \f$c\f$. 
 *
 *
 *	\par Space complexity
 *		\f$\Order{n\log|\Sigma| + 2|\Sigma|\log n}\f$ bits, where \f$n\f$ is the size of the vector the wavelet tree was build for.
 *
 *   The wavelet tree was proposed first by Grossi et al. 2003 and applied to the BWT in Foschini et al. 2004.
 */
template<class RandomAccessContainer=unsigned char*, class RankSupport = rank_support_v<>, class SelectSupport=select_support_mcl<>, class SelectSupportZero=select_support_mcl<0> >
class wavelet_tree{
	public:
	typedef typename wavelet_tree_trait<RandomAccessContainer>::size_type 		size_type;
	typedef typename wavelet_tree_trait<RandomAccessContainer>::value_type 		value_type;
	typedef typename wavelet_tree_trait<RandomAccessContainer>::map_type		map_type;
	typedef typename wavelet_tree_trait<RandomAccessContainer>::inv_map_type	inv_map_type;
	private:
		size_type 			m_size;
		size_type 			m_sigma; 		//<- \f$ |\Sigma| \f$
		bit_vector 			m_tree;			// bit vector to store the wavelet tree
		RankSupport			m_tree_rank;	// rank support for the wavelet tree bit vector
		SelectSupport		m_tree_select1;	// select support for the wavelet tree bit vector
		SelectSupportZero	m_tree_select0; 
		int_vector<64>		m_node_pointers;
		int_vector<64>		m_node_pointers_rank;

		value_type			m_first_symbol;

		map_type			m_char_map;		// map the characters to integers
		inv_map_type		m_inv_char_map;	// map the integers back to characters

		void copy(const wavelet_tree &wt){
			m_size 			= wt.m_size;
			m_sigma 		= wt.m_sigma;
			m_tree			= wt.m_tree;
			m_tree_rank 	= wt.m_tree_rank;
			m_tree_rank.set_vector(&m_tree);
			/*
			m_tree_select1	= wt.m_tree_select1;
			m_tree_select1.set_vector(&m_tree);
			m_tree_select0	= wt.m_tree_select0;
			m_tree_select0.set_vector(&m_tree);
			*/
			m_node_pointers	= wt.m_node_pointers;
			m_node_pointers_rank = wt.m_node_pointers_rank;
			m_first_symbol	= wt.m_first_symbol;
			m_char_map		= wt.m_char_map;
			m_inv_char_map	= wt.m_inv_char_map;
		}	

	public:	
	
	const size_type &sigma;	

	// Default constructor
	wavelet_tree():m_size(0),m_sigma(0), sigma(m_sigma){};

	//! Constructor
	/*!
	 *	\param rac Reference to the vector (or unsigned char array) for which the wavelet tree should be build.  
	 *	\param size Size of the prefix of the vector (or unsigned char array) for which the wavelet tree should be build.
	 *	\par Time complexity
	 *		\f$ \Order{n\log|\Sigma|}\f$, where \f$n=size\f$
	 */
	wavelet_tree(const typename wavelet_tree_trait<RandomAccessContainer>::reference_type rac, size_type size):m_size(size), m_sigma(0), sigma(m_sigma){
#ifdef SDSL_DEBUG_WAVELET_TREE	
		std::cerr<<"Wavelet Tree construct: size="<<size<<std::endl;
#endif		
		// calculate alphabet size and the mappings for the symbols to the integers and back
		m_sigma = wavelet_tree_trait<RandomAccessContainer>::alphabet_size_and_map(rac, m_size, m_char_map, m_inv_char_map, m_first_symbol);

		int_vector<> node_sizes = int_vector<>(2*m_sigma+1, 0, bit_magic::l1BP(m_size)+1);
		m_node_pointers = int_vector<64>(node_sizes.size()+1, 0);
		m_node_pointers_rank = int_vector<64>(node_sizes.size()+1, 0);

		if( m_sigma < 2 ){
			return;
		}else{
			for(size_type i=0; i<m_size; ++i){
				size_type lex_idx = m_char_map[rac[i]]; // lex_idx in [0..m_sigma-1]
				size_type sigma = m_sigma, node=0;
				while( sigma >= 2 ){
#ifdef SDSL_DEBUG_WAVELET_TREE					
					if(node>=node_sizes.size()) {
						std::cerr<<" node="<<node<<" node_sizes.size()="<<node_sizes.size()<<" rac["<<i<<"]="<<rac[i]<<" orig_lex_idx="<< (int)m_char_map[rac[i]] <<" sigma="<<sigma<<" lex_idx="<<lex_idx<<std::endl;
					}
#endif					
					assert(node<node_sizes.size());
					node_sizes[node] = node_sizes[node]+1;
					if( lex_idx < (sigma+1)/2 ){
						sigma = (sigma+1)/2;
						node = 2*node+1;
					}else{
						lex_idx -= (sigma+1)/2;
						sigma -= (sigma+1)/2;
						node = 2*node+2;
					}
				}
				if(lex_idx!=0){
					std::cout<<" _"<<rac[i]<<" ("<<lex_idx<<")";
				}
			}	
#ifdef SDSL_DEBUG_WAVELET_TREE			
			std::cerr<<"wavelet tree node sizes"<<std::endl;
			for(size_type i=0; i<node_sizes.size(); ++i){ std::cerr<<node_sizes[i]<<" "; } std::cerr << std::endl;
#endif

			size_type max_node = 0;
			m_node_pointers[0] = 0;
			for(size_type i=1; i < m_node_pointers.size(); ++i){ 
				m_node_pointers[i] = m_node_pointers[i-1] + node_sizes[i-1];
				if( node_sizes[i-1]!=0 )
					max_node = i-1;
			}
			node_sizes = int_vector<>(max_node+1, 0, 64);
#ifdef SDSL_DEBUG_WAVELET_TREE
			std::cerr<<"all nodes size = "<<m_node_pointers[max_node+1]<<std::endl;
#endif	
			m_tree = bit_vector( m_node_pointers[max_node+1] );
			for(size_type i=0; i<m_size; ++i){
				size_type lex_idx = m_char_map[rac[i]]; // lex_idx in [0..m_sigma-1]
				size_type sigma = m_sigma, node=0;
				while( sigma >= 2 ){
					m_tree[m_node_pointers[node]+node_sizes[node]] = (lex_idx >= (sigma+1)/2);
					node_sizes[node] = node_sizes[node]+1;
					if( lex_idx < (sigma+1)/2 ){
						sigma = (sigma+1)/2;
						node = 2*node+1;
					}else{
						lex_idx -= (sigma+1)/2;
						sigma -= (sigma+1)/2;
						node = 2*node+2;
					}
				}
			}			
			m_tree_rank.init(&m_tree);
			for(size_type i=0; i < m_node_pointers.size(); ++i){
				m_node_pointers_rank[i] = m_tree_rank( m_node_pointers[i] );
			}
//			m_tree_select0.init(&m_tree);
//			m_tree_select1.init(&m_tree);
		}
	}

	//! Copy constructor
	wavelet_tree(const wavelet_tree &wt):sigma(m_sigma){
		copy(wt);
	}	

	//! Assignment operator
	wavelet_tree& operator=(const wavelet_tree &wt){
		if( this != &wt ){
			copy(wt);
		}
		return *this;
	}

	//! Swap operator
	void swap(wavelet_tree &wt){
		if(this != &wt){
			std::swap(m_size, wt.m_size);
			std::swap(m_sigma,  wt.m_sigma);
			m_tree.swap(wt.m_tree);
			m_tree_rank.swap(wt.m_tree_rank); // rank swap after the swap of the bit vector m_tree
			m_tree_select1.swap(wt.m_tree_select1); // select1 swap after the swap of the bit vector m_tree
			m_tree_select0.swap(wt.m_tree_select0); // select0 swap after the swap of the bit vector m_tree
			m_node_pointers.swap(wt.m_node_pointers);
			m_node_pointers_rank.swap(wt.m_node_pointers_rank);
			std::swap(m_first_symbol, wt.m_first_symbol);
			m_char_map.swap(wt.m_char_map);
			m_inv_char_map.swap(wt.m_inv_char_map);
		}
	}

	//! Returns the size of the original vector.
	size_type size()const{
		return m_size;
	}

	//! Returns whether the wavelet tree contains no data.
	bool empty()const{
		return m_size == 0;
	}

	//! Recovers the ith symbol of the original vector.
	/*! \param i The index of the symbol in the original vector. \f$i \in [0..size()-1]\f$
	 *	\returns The ith symbol of the original vector. 
	 */
	value_type operator[](size_type i)const{
		size_type lex_idx	= 0;
		size_type sigma		= m_sigma;
		size_type node		= 0;
		while( sigma >= 2 ){
			if( m_tree[ m_node_pointers[node]+i ] ){ // go to the right child
				lex_idx += (sigma+1)/2;
				i = m_tree_rank( m_node_pointers[node]+i ) - m_node_pointers_rank[node];//m_tree_rank( m_node_pointers[node] );
				node = 2*node+2;
				sigma -= (sigma+1)/2;
			}else{ // go to the left child
				i = i - (m_tree_rank( m_node_pointers[node]+i ) - m_node_pointers_rank[node]); //m_tree_rank( m_node_pointers[node] ) );
				node = 2*node+1;
				sigma = (sigma+1)/2;
			}
		}
		return m_inv_char_map[lex_idx];
	};

/*!
	 *  \param i The exclusive index of the prefix range [0..i-1], so \f$i\in[0..size()]\f$.
   	 *  \param c The symbol to count the occurences in the prefix.
	 *	\returns -1 if the symbol at position i is not c, and otherwise the number of occurences of symbol c in the prefix [0..i-1] of the supported vector. 
	 *  \par Time complexity
	 *		\f$ \Order{\log |\Sigma|} \f$
	 */
	int rank_special(size_type i, value_type c)const {
		size_type sigma		= m_sigma;
		size_type lex_idx	= m_char_map[c];
		size_type node		= 0;
		while( sigma >= 2 ){
			if( m_tree[ m_node_pointers[node]+i ] ){ // go to the right child
				i = m_tree_rank( m_node_pointers[node]+i ) - m_node_pointers_rank[node];
				node = 2*node+2;
				if (lex_idx < (sigma+1)/2)
					return -1;
				lex_idx -= (sigma+1)/2;
				sigma /= 2;
			}else{ // go to the left child
				i = i - (m_tree_rank( m_node_pointers[node]+i ) - m_node_pointers_rank[node]);
				node = 2*node+1;
				sigma = (sigma+1)/2;
				if (lex_idx >= sigma)
					return -1;
			}
		}
		return i;
	}


	//TODO: buchstaben, die nicht im orginaltext vorkommen behandeln!!! DONE: jetzt testen
	//! Calculates how many symbols c are in the prefix [0..i-1] of the supported vector.
	/*!
	 *  \param i The exclusive index of the prefix range [0..i-1], so \f$i\in[0..size()]\f$.
   	 *  \param c The symbol to count the occurences in the prefix.
	 *	\returns The number of occurences of symbol c in the prefix [0..i-1] of the supported vector. 
	 *  \par Time complexity
	 *		\f$ \Order{\log |\Sigma|} \f$
	 */
	size_type rank(size_type i, value_type c)const{
		assert(i>=0 and i <= size());
		if( !wavelet_tree_trait<RandomAccessContainer>::symbol_available(m_char_map, c, m_first_symbol) )
			return 0;
		size_type lex_idx 	= m_char_map[c];
		size_type sigma 	= m_sigma;
		size_type node		= 0;
		size_type result	= i;
		while( sigma >= 2 and result > 0 ){
			if( lex_idx < (sigma+1)/2 ){
				result 	= result - (m_tree_rank(m_node_pointers[node]+result) -  m_node_pointers_rank[node]); //m_tree_rank(m_node_pointers[node]));
				sigma 	= (sigma+1)/2;
				node	= 2*node+1;
			}else{
				result = m_tree_rank(m_node_pointers[node]+result) -  m_node_pointers_rank[node]; // m_tree_rank(m_node_pointers[node]);
				lex_idx -= (sigma+1)/2;
				sigma -= (sigma+1)/2;
				node = 2*node+2;
			}
		}
		return result;
	};

	//! Calculates the ith occurence of the symbol c in the supported vector.
	/*!
	 *  \param i The ith occurence. \f$i\in [1..rank(size(),c)]\f$.
	 *  \param c The symbol c.
	 *  \par Time complexity
	 *		\f$ \Order{\log |\Sigma|} \f$
	 */
	 // TODO: was ist wenn c gar nicht vorkommt, oder es keine i vorkommen gibt?
	size_type select(size_type i, value_type c)const{
		assert(i > 0);
#ifdef SDSL_DEBUG_WAVELET_TREE
		std::cerr<<"wavelet_tree: select("<<i<<", "<<c<<")"<<std::endl;
		std::cerr<<" i="<<i<<std::endl;
		std::cerr<<"c=\""<<c<<"\""<<std::endl;
		std::cerr<<"size()="<<size()<<std::endl;
		std::cerr<<rank(size(),c)<<std::endl;
#endif		
	    assert(i <= rank(size(), c) );
		size_type lex_idx 	= m_char_map[c];
		size_type sigma 	= m_sigma;
		size_type node		= 0;
		// first phase: go down to the node corresponding to the lex_idx of the character
		while( sigma >= 2 ){
			if( lex_idx < (sigma+1)/2 ){
				sigma = (sigma+1)/2;	
				node = 2*node+1;
			}else{
				lex_idx -= (sigma+1)/2;
				sigma -= (sigma+1)/2;
				node = 2*node+2;
			}
		}
		// second phase go up and select the right position
		size_type result = i-1;
		while(node != 0){
			if( node&1 ){// node is a left child
				node = (node-1)/2;
				size_type node_idx = m_node_pointers[node];
				result = m_tree_select0(node_idx-m_tree_rank(node_idx)+result+1)-node_idx;
			}else{//node is a right child
				node = (node-1)/2;
				result = m_tree_select1(m_tree_rank(m_node_pointers[node])+result+1)-m_node_pointers[node];
			}
		}
		return result;
	};

	//! Serializes the data structure into the given ostream
	size_type serialize(std::ostream &out)const{
		size_type written_bytes = 0;
		out.write((char*)&m_size, sizeof(m_size));
		written_bytes += sizeof(m_size);
		out.write((char*)&m_sigma, sizeof(m_sigma));
		written_bytes += sizeof(m_sigma);
		written_bytes += m_tree.serialize(out);
		written_bytes += m_tree_rank.serialize(out);
		written_bytes += m_tree_select1.serialize(out);
		written_bytes += m_tree_select0.serialize(out);
		written_bytes += m_node_pointers.serialize(out);
		written_bytes += m_node_pointers_rank.serialize(out);
		written_bytes += wavelet_tree_trait<RandomAccessContainer>::serialize_maps(out, m_char_map, m_inv_char_map);
		return written_bytes;
	}

	//! Loads the data structure from the given istream.
	void load(std::istream &in){
		in.read((char*) &m_size, sizeof(m_size));
		in.read((char*) &m_sigma, sizeof(m_sigma));
		m_tree.load(in);
		m_tree_rank.load(in, &m_tree);
		m_tree_select1.load(in, &m_tree);
		m_tree_select0.load(in, &m_tree);
		m_node_pointers.load(in);
		m_node_pointers_rank.load(in);
		wavelet_tree_trait<RandomAccessContainer>::load_maps(in, m_char_map, m_inv_char_map);
	}
/*
	void print_info()const{
		size_type rle_ed = 0;
		for(size_type i=0; i < m_tree.size(); ++i){
			
		}
	}
*/	
	
};	

}// end namespace sds

#endif // end file 
