#include "sdsl/suffixtrees.hpp"
#include "sdsl/algorithms.hpp"
#include "sdsl/lcp_uncompressed.hpp"
#include "sdsl/util.hpp"
#include "sdsl/testutils.hpp"

#include <iostream>
#include <vector>
#include <algorithm>
#include <list>
#include <string>
#include <string.h>
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <utility>
#include <iomanip>
#include <math.h>
#include <sys/time.h>
#include <time.h>

using namespace sdsl;
using namespace std;

typedef csa_wt<csa_wt<>::wavelet_tree_type, 10000, 10000> tCSA_WT;
typedef cst_sct2< tCSA_WT, lcp_kurtz<tCSA_WT> > tCST;
typedef tCST::size_type size_type;

bool backward_search(
		const tCST::csa_type &csa,
		const unsigned char c,
		tCST::node_type &p){
	typedef tCST::size_type size_type;
	size_type cc = csa.char2comp[c];
	if (!cc)
		return false;
	size_type c_before_r = csa.wavelet_tree.rank(p.right, c);
	if (!c_before_r)
		return false;
	size_type c_before_l = csa.wavelet_tree.rank(p.left, c);
	if (c_before_l == c_before_r)
		return false;
	p.left = csa.C[cc]+c_before_l;
	p.right = csa.C[cc]+c_before_r;
	return true;
}

double param;

inline double constantWeight(double parent_lcp, double length) {
	return length - parent_lcp;
}

inline double linearWeight(double parent_lcp, double length) {
	return (length * (length + 1))/2 - (parent_lcp * (parent_lcp + 1))/2;
}

inline double expDecayWeight(double parent_lcp, double length) {
	if (parent_lcp == length)
		return 0;
	return (pow(param,-parent_lcp)-pow(param,-length))/(param-1);
}

inline double kSpectrumWeight(double parent_lcp, double length) {
	if (length >= param && parent_lcp < param)
		return 1;
	return 0;
}

inline double boundedRangeWeight(double parent_lcp, double length) {
	return max(0.0, min(length, param) - parent_lcp);
}

double (*getWeight)(double,double) = NULL;

float *alpha_ps;

float computeK(const tCST &cst, float *valtab, const unsigned char *s, int length) {
	int l1, l2, ind;
	float res = 0;
	tCST::node_type p(0, 0, cst.size());
	// calculate matching statistics
	for (int p2 = length-1; p2 >= 0; --p2)
		if( backward_search(cst.csa, s[p2], p) ){
			++p.l;
			l1 = cst.lcp[p.left];
			l2 = (p.right < cst.size()? cst.lcp[p.right] : -1);
			if (l1 < l2) {
				ind = p.right;
				l1 = l2;
			}
			else
				ind = p.left;
			res += valtab[ind] + getWeight(l1, p.l) * (alpha_ps[p.right]-alpha_ps[p.left]);
		}else if (p.l > 0){
			--p.right;
			p = cst.parent(p);
			++p.right;
			++p2;
		}
	return res;
}

void compute_valtab(const tCST &cst1, int i, int j, int l, int idx, double val, float *valtab) {
	int lb = i, m;
	do {
		valtab[idx] = val;
		if (lb < idx-1) {
			size_type t = cst1.bp_support.select(idx+1)-1;
			size_type k = cst1.bp_support.rank(cst1.bp_support.find_open(t))-1;
			m = cst1.lcp[k];
			compute_valtab(cst1, lb, idx-1, m, k, val + getWeight(l, m) * (alpha_ps[idx] - alpha_ps[lb]), valtab);
		}
		if (idx == j)
			return;
		lb = idx;
		size_type t = cst1.bp_support.find_close(cst1.bp_support.select(idx+1))-1;
		idx = cst1.bp_support.rank(cst1.bp_support.find_open(t)) - 1;
	} while((m = cst1.lcp[idx]) == l);
	compute_valtab(cst1, lb, j, m, idx, val + getWeight(l, m) * (j - lb + 1), valtab);
}

void exit_error(const string &message) {
	cout << message << endl;
	cout << "usage: ./predict <weighting function> <wf parameter> <support vector file> <dataset filename>" << endl;
	cout << "<weighting function> and <wf parameter> is one of:" << endl;
	cout << "0 - constant weight (dummy parameter)" << endl;
	cout << "1 - exponential decay (parameter lamda)" << endl;
	cout << "2 - K-Spectrum (parameter k)" << endl;
	cout << "3 - Bounded Range (parameter n)" << endl;
	cout << "4 - linear weight (dummy parameter)" << endl;
	cout << "the i-th line of the support vector file should contain a triple of <filename of text file containing support vector i> <alpha_i> <y_i>" << endl;
	cout << "every line of the dataset is considered as one filename containing a string for which a prediction should be calculated" << endl;
	exit(1);
}

char *readFile(const string &filename, size_type &size) {
	ifstream text_fp(filename.c_str(), ios::in|ios::ate);
	if (!text_fp)
		exit_error(string("Error opening text file ") + filename);
	size = text_fp.tellg();
	char *text = new char[size + 1];
	text_fp.seekg(0, ios::beg);
	text_fp.read(text, size);
	text_fp.close();
	for (size_type i=0; i<size; ++i)
		if (text[i] == '\n')
			text[i] = ' ';
	text[size] = 0;
	return text;
}

int main(int argc, char **argv){
	if (argc != 5)
		exit_error("expected 5 parameter");
	int type;
	if (sscanf(argv[1], "%d", &type) != 1 || type < 0 || type > 4)
		exit_error("weighting function must be an integer between 0 and 4");
	switch(type) {
		case 0: getWeight = constantWeight; break;
		case 1: if (sscanf(argv[2], "%lf", &param) != 1) exit_error("expected a wf parameter lamda"); getWeight = expDecayWeight; break;
		case 2: if (sscanf(argv[3], "%lf", &param) != 1) exit_error("expected a wf parameter k"); getWeight = kSpectrumWeight; break;
		case 3: if (sscanf(argv[4], "%lf", &param) != 1) exit_error("expected a wf parameter n"); getWeight = boundedRangeWeight;
		case 4: getWeight = linearWeight; break;
	}
	ifstream supportvectors(argv[3]);
	if (!supportvectors)
		exit_error("Error opening support vector file");
	string line;
	string text;
	string filename;
	double alpha;
	int y;
	vector<double> alpha_y;
	vector<size_type> *start = new vector<size_type>;
	while(getline(supportvectors, line)) {
		istringstream istr(line);
		if (!(istr >> filename >> alpha >> y)) {
			char buf[1000];
			sprintf(buf, "wrong format in line %d of support vector file", (int)alpha_y.size());
			exit_error(buf);
		}
		alpha_y.push_back(y * alpha);
		size_type size;
		if (text.size() > 0)
			text += (char)1; // separating symbol
		char *s = readFile(filename, size);
		start->push_back(text.size());
		text += s;
		delete [] s;
	}
	bit_vector *bp = new bit_vector(text.size()+1, 0);
	// now construct suffix array of the text
	for (vector<size_type>::iterator it=start->begin(); it!=start->end(); ++it)
		(*bp)[*it] = 1;
	delete start;
	rank_support_v<1,1> *rs = new rank_support_v<1,1>(bp);
	csa_uncompressed *csa_un = new csa_uncompressed((const unsigned char *)text.c_str());
	alpha_ps = new float[csa_un->size() + 1];
	alpha_ps[0] = alpha_ps[1] = 0;
	for (int i=1; i<(int)csa_un->size(); ++i) {
		// get number of support vector for current suffix
		int support_vector_index = rs->rank((*csa_un)[i] + 1) - 1;
		assert(support_vector_index >= 0 && support_vector_index < (int)alpha_y.size());
		alpha_ps[i + 1] = alpha_ps[i] + alpha_y[support_vector_index];
	}
	delete rs;
	delete bp;
	tCST cst((const csa_uncompressed)*csa_un, (const unsigned char *)text.c_str());
	delete csa_un;
	text.clear();

	float *valtab = new float[cst.size()];
	compute_valtab(cst, 0, cst.size()-1, 0, 1, 0, valtab);
	ifstream dataset(argv[4]);
	if (!dataset)
		exit_error("Error opening dataset file");
	while(getline(dataset, filename)) {
		size_type size;
		char *pattern = readFile(filename, size);
		double gs = computeK(cst, valtab, (const unsigned char *)pattern, size);
		cout << gs << endl;
		delete [] pattern;
	}
	delete [] valtab;
	delete [] alpha_ps;
	return 0;
}
