#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/balanced_parentheses_support.hpp>
#include <sdsl/util.hpp>
#include <sdsl/rank_support_v5.hpp>
#include <stack>
#include <vector>
#include <string>

typedef sdsl::bit_vector::size_type size_type;
typedef sdsl::balanced_parentheses_support_simple<sdsl::balanced_parentheses_support_simple<>::nnd_type, sdsl::rank_support_v5<> > balanced_parentheses_support_type;

using namespace boost::unit_test;
using namespace boost::test_tools;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( balanced_parentheses_support_j_test_suite )

void create_random_tree(size_type begin, size_type end, size_type max_degree, bit_vector &v){
	if(begin>=end)
		return;
//	BOOST_MESSAGE("begin = "<<begin<<" end = "<<end);
	size_type interval_size = end-begin;
	assert(interval_size%2==0 and interval_size>1);
	size_type degree = ((rand())%max_degree)+1;
	if(degree*2 > interval_size)
		degree = interval_size/2;
	for(size_type i=0; i<degree-1; ++i){
		interval_size = end-begin;
		size_type size = rand()%(interval_size+1-2*(degree-i-1));
		if(size==0)
			size=2;
		if(size%2==1)
			++size;
		v[begin] = 1;
		v[begin+size-1] = 0;
		create_random_tree(begin+1, begin+size-1, max_degree, v);
		begin = begin+size;
	}
	v[begin] = 1;
	v[end-1] = 0;
	create_random_tree(begin+1, end-1, max_degree, v);
}

BOOST_AUTO_TEST_CASE( correctness_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_MESSAGE( "Running correctness test for balanced_parentheses_support_j..." );
	const size_type test_cases = 8;
	bit_vector v[test_cases];
	uint32_t block_size[test_cases] = {0};
	
	v[0].resize(10000000);
	create_random_tree(0, v[0].size(), 4, v[0]);
	BOOST_MESSAGE("test case 1: random tree of size "<<v[0].size()<<" and max branching factor "<<4<<" was created");
	v[1].resize(10000000);
	create_random_tree(0, v[1].size(), 1, v[1]);
	BOOST_MESSAGE("test case 2: random tree of size "<<v[1].size()<<" and max branching factor "<<1<<" was created");
	v[2].resize(10000000);
	create_random_tree(0, v[2].size(), 26, v[2]);
	BOOST_MESSAGE("test case 3: random tree of size "<<v[2].size()<<" and max branching factor "<<26<<" was created");
	v[3].resize(10000000);
	create_random_tree(0, v[3].size(), 128, v[3]);
	BOOST_MESSAGE("test case 4: random tree of size "<<v[3].size()<<" and max branching factor "<<128<<" was created");
	v[4] = bit_vector(18,0);
	v[4][0]=v[4][1]=v[4][2]=v[4][4]=v[4][5]=v[4][9]=v[4][10]=v[4][12]=v[4][13]=1;
	BOOST_MESSAGE("test case 5: hand crafted tree of size 18");

	v[5].resize(10000000);
	for(size_type i=0; i<v[5].size(); i+=2){
		v[5][i] = 1;
		v[5][i+1] = 0;
	}
	BOOST_MESSAGE("test case 6: tree consisting of "<<v[5].size()/2<<" roots");

	v[6] = bit_vector(10,0);
	v[6][0]=v[6][1]=v[6][4]=v[6][6]=v[6][7]=1;
	block_size[6] = 5;
	BOOST_MESSAGE("test case 7: hand crafted tree of size 10");
	balanced_parentheses_support_type bps(&v[6], block_size[6]);
	size_type rr_enc = bps.rr_enclose(1,7);
	if( rr_enc != 6 ){
		BOOST_MESSAGE(" rr_enclose(1,7)="<<rr_enc<<" expected="<<6);
	}	
	BOOST_REQUIRE( bps.rr_enclose(1,7) == 6 );

	v[7] = bit_vector(16,0);
	v[7][0]=v[7][1]=v[7][3]=v[7][5]=v[7][7]=v[7][9]=v[7][11]=v[7][12]=1;
	block_size[7] = 5;
	BOOST_MESSAGE("test case 7: hand crafted tree of size 10");
	bps = balanced_parentheses_support_type(&v[7], block_size[7]);
	rr_enc = bps.rr_enclose(1,12);
	if( rr_enc != 11 ){
		BOOST_MESSAGE(" rr_enclose(1,12)="<<rr_enc<<" expected="<<11);
	}
	BOOST_REQUIRE( bps.rr_enclose(1,12) == 11 );

	for(size_type tc=0; tc<test_cases; ++tc){
		BOOST_MESSAGE("test case "<<tc+1);
		std::stack<size_type> s;
		std::vector<size_type> rank(v[tc].size(),0);
		std::vector<size_type> excess(v[tc].size(),0);
		std::vector<size_type> matches(v[tc].size(),0);
		std::vector<size_type> enclose(v[tc].size(),0);
		// calculate rank and select values
		for(size_type i=0, rank_value=0; i < v[tc].size(); ++i){
			if(v[tc][i]){
				++rank_value;
			}
			rank[i] = rank_value;
		}
		// calculate excess answers
		for(size_type i=0, excess_val=0; i < v[tc].size(); ++i){
			if(v[tc][i])
				++excess_val;
			else
				--excess_val;
			excess[i] = excess_val;
		}
		BOOST_REQUIRE(s.empty());
		// calculate find_open and find_close answers
		for(size_type i=0; i < v[tc].size(); ++i){
			if(v[tc][i])
				s.push(i);
			else{
				BOOST_REQUIRE(!s.empty());
				matches[s.top()] = i;
				matches[i] = s.top();
				s.pop();
			}
		}
		BOOST_REQUIRE(s.empty());
		// calculate enclose answers
		for(size_type i=0; i < v[tc].size(); ++i){
			if(v[tc][i]){ // found an opening parenthesis
				if(s.empty())
					enclose[i] = v[tc].size();// there is no enclosing pair for the current parenthesis
				else
					enclose[i] = s.top();
				s.push(i);
			}
			else{// found a closing parenthesis
				BOOST_REQUIRE(!s.empty());
				enclose[i] = enclose[s.top()];// s.top() == find_open(i)	
				s.pop();
			}
		}
		BOOST_REQUIRE(s.empty());

		// create balanced_parentheses_support for v
		BOOST_MESSAGE("Creating balanced_parentheses_support_j");
		balanced_parentheses_support_type bp_support;//(&v[tc], 64);
		BOOST_MESSAGE("Creating ...");
		bp_support = balanced_parentheses_support_type(&v[tc], block_size[tc] ? block_size[tc] : 6 );
		
		bool tested_serialize_and_load = false;
test_methods:		
		size_type size_of_bp_support_in_bytes = util::get_size_in_bytes(bp_support); 
		BOOST_MESSAGE("Size of bp_support in bytes = "<<size_of_bp_support_in_bytes<<" for "<<(bp_support.size()/8)<<" bytes of bp data.");
		// check rank values
///*		
		BOOST_MESSAGE("Check rank and select values");
		for(size_type i=0; i < v[tc].size(); ++i){
			BOOST_REQUIRE( rank[i] == bp_support.rank(i) );
		}
		for(size_type i=0, cnt=0; i < v[tc].size(); ++i){
			if(v[tc][i]){
				++cnt;
				BOOST_REQUIRE( bp_support.select(cnt)==i  );
			}
		}
		// check excess answers
		BOOST_MESSAGE("Check excess answers");
		for(size_type i=0; i < v[tc].size(); ++i){
			BOOST_REQUIRE( excess[i] == bp_support.excess(i) );
		}
		// check find_close
		BOOST_MESSAGE("Check find_close answers");
		for(size_type i=0; i < v[tc].size(); ++i){
//			BOOST_MESSAGE(" i = "<<i);
			size_type res = bp_support.find_close(i);
			if(v[tc][i]==1){
				if(res!=matches[i]){
					BOOST_MESSAGE("i = "<<i<<" res = "<<res<<" match[i] = "<<matches[i]<<" excess["<<matches[i]<<"]="<<excess[matches[i]]<<" excess["<<res<<"]="<<excess[res]);
				}
				BOOST_REQUIRE( res == matches[i] );
			}
		}
		// check find_open
		BOOST_MESSAGE("Check find_open answers");
		for(size_type i=0; i < v[tc].size(); ++i){
//		for(size_type i=v[tc].size()-1; i > 0; --i){
			if(i % 1000000 == 0)
			BOOST_MESSAGE(i);
			size_type res = bp_support.find_open(i);
			if(v[tc][i]==0){
				if(res!=matches[i]){
					BOOST_MESSAGE("i = "<<i<<" res = "<<res<<" match[i] = "<<matches[i]<<" excess["<<matches[i]<<"]="<<excess[matches[i]]<<" excess["<<res<<"]="<<excess[res]);
					BOOST_MESSAGE("excess[21]="<<excess[21]);
				}
				BOOST_REQUIRE( res == matches[i] );
			}
		}

		// check enclose 
		BOOST_MESSAGE("Check enclose answers");
		size_type answered = 0;
		for(size_type i=0; i < v[tc].size(); ++i){
//			BOOST_MESSAGE(i);
			if(v[tc][i]==1){
				size_type res = bp_support.enclose(i);
				if(res!=v[tc].size()){
					if(res!=enclose[i]){
						BOOST_MESSAGE("i = "<<i<<" res = "<<res<<" enclose[i] = "<<enclose[i]);
					}
					BOOST_REQUIRE( res == enclose[i] );
					++answered;
				}
			}
		}
		BOOST_MESSAGE("answered "<<((double)answered)/(v[tc].size()/2));
//*/		
//		continue;

		if(true){
			// check restricted enclose
			size_type rr_answers = 0;
			size_type rr2_answers = 0;
			BOOST_MESSAGE("Check restricted enclose");
			for(size_type i=0; i < v[tc].size(); ++i){
				if(v[tc][i]==1){
					size_type mi = bp_support.find_close(i);
					size_type opening_parentheses_so_far =  bp_support.rank( mi );
					if( opening_parentheses_so_far < v[tc].size()/2 ){ // choose k
						opening_parentheses_so_far += 1;// + (rand() % remaining_opening_parentheses);
						size_type k = bp_support.select( opening_parentheses_so_far );
						
						if( opening_parentheses_so_far < v[tc].size()/2 ){ // choose j
							size_type mk = bp_support.find_close(k);
							if(mk > k+1){ // if k is not a leave
								size_type j = bp_support.select( opening_parentheses_so_far + 1 + rand()%((mk-k)/2) );
								size_type mj = bp_support.find_close(j);
								
								size_type res = bp_support.rr_enclose_naive(i, j);
								size_type res2 = bp_support.rr_enclose(i, j);
								if( res != k ){
									size_type mres = bp_support.find_close(res);
									BOOST_MESSAGE("i="<<i<<" mi="<<mi<<" j="<<j<<" mj="<<mj<<" k="<<k<<" restricted_enclose(i,j)="<<res<<" "<<mres);
								}
								BOOST_REQUIRE(res == k);
								++rr_answers;
								if( res2 != v[tc].size() or res == v[tc].size() ){
									++rr2_answers;
									if( res2 != res ){
										BOOST_MESSAGE("i="<<i<<" mi="<<mi<<" j="<<j<<" mj="<<mj<<" k="<<k<<" restricted_enclose(i,j)="<<res<<" rr_enclose(i,j)="<<res2);
									}
									BOOST_REQUIRE(res2 == res);
								}
							}
						}
					}
				}
			}
			if(rr_answers)
				BOOST_MESSAGE("answered "<<((double)rr2_answers)/(rr_answers));
			// check preceding_closing_parentheses
			BOOST_MESSAGE("Check preceding_closing_parentheses");
			for(size_type i=0, preceding_zeros=0; i < v[tc].size(); ++i){
				if(v[tc][i]==1){
					BOOST_REQUIRE( bp_support.preceding_closing_parentheses(i) == preceding_zeros );
					preceding_zeros = 0;
				}else{
					++preceding_zeros;
				}
			}
	
		}

		if( tested_serialize_and_load )
			continue;	

		BOOST_MESSAGE("Check serialize");
		std::string file_name = "/tmp/test_bp"; 
		util::store_to_file(bp_support, file_name.c_str() );

		

		BOOST_MESSAGE("Check load");
		std::ifstream in;
		in.open(file_name.c_str(), std::ios::binary | std::ios::in );
		BOOST_REQUIRE( in );
		bp_support.load(in, &v[tc]);
		in.close();

		tested_serialize_and_load = true;
		goto test_methods; 

	}
}

BOOST_AUTO_TEST_SUITE_END()
