#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/coder.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/testutils.hpp> 
#include <ctime> // for performance measurement


using namespace boost::unit_test;
using namespace boost::test_tools;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( coding_test_suite )

BOOST_AUTO_TEST_CASE( coding_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_MESSAGE( "Running correctness test for coder..." );

	int_vector<> v(10), z, w;
	for(size_t i=0; i<v.size();++i){
		v[i] = 0xFFFFFFFFFFFFFFFFULL-i;
		BOOST_CHECK(v[i] == 0xFFFFFFFFFFFFFFFFULL-i);
	}
	coder::fibonacci::encode(v, z);		
	coder::fibonacci::decode(z, w);
	BOOST_CHECK( v.size() == w.size() );
	BOOST_CHECK( v == w );
	coder::elias_delta::encode(v, z);		
	coder::elias_delta::decode(z, w);
	BOOST_CHECK( v.size() == w.size() );
	BOOST_CHECK( v == w );
	coder::ternary::encode(v,z);
	coder::ternary::decode(z,w);
	BOOST_CHECK( v.size() == w.size() );
	BOOST_CHECK( v == w );

	uint8_t intWidth = 33;// <=63 !!!
	size_t  size = 100000;
	v.setIntWidth(intWidth);
	v.resize( size );

	for(size_t i = 0; i < v.size(); ++i)
		v[i] = i+1;

	BOOST_MESSAGE( "encode_fibonacci..." );
	coder::fibonacci::encode(v, z);
	
	BOOST_MESSAGE( "decode_fibonacci..." );
	coder::fibonacci::decode(z, w);

	bool are_equal = true;
	size_t i = 0;
	BOOST_CHECK_EQUAL( v.bit_size(), w.bit_size() );
	BOOST_CHECK_EQUAL( v.getIntWidth(), w.getIntWidth() );

	for(; i < v.size() && are_equal; ++i)
		are_equal  = (v[i]==w[i]);
	BOOST_CHECK_MESSAGE( are_equal, "contents of v and w differ at position " << i );
	if(!are_equal and intWidth * size < 1000){
		v.setIntWidth(1); z.setIntWidth(1); w.setIntWidth(1);
		BOOST_MESSAGE("v="<<v);	
		BOOST_MESSAGE("z="<<z);	
		BOOST_MESSAGE("w="<<w);	
		v.setIntWidth(intWidth); z.setIntWidth(intWidth); w.setIntWidth(intWidth);
	}

	for(size_t i = 0; i < v.size(); ++i)
		v[i] = (1ULL << v.getIntWidth())-1;

	coder::fibonacci::encode(v, z);
	coder::fibonacci::decode(z,w);

	are_equal = true;
	i = 0;
	BOOST_CHECK_EQUAL( v.bit_size(), w.bit_size() );
	BOOST_CHECK_EQUAL( v.getIntWidth(), w.getIntWidth() );
	for(; i < v.size() && are_equal; ++i)
		are_equal  = (v[i]==w[i]);
	BOOST_CHECK_MESSAGE( are_equal, "contents of v and w differ at position " << i );

	BOOST_MESSAGE( "encode_elias_delta..." );
	coder::elias_delta::encode(v,z);
	BOOST_MESSAGE( "decode_elias_delta..." );
	coder::elias_delta::decode(z,w);
	
	are_equal  = true;
	i=0;
	BOOST_CHECK_EQUAL( v.size(), w.size() );
	BOOST_CHECK_EQUAL( v.getIntWidth(), w.getIntWidth() );
	for(; i < v.size() && are_equal; ++i)
		are_equal  = (v[i]==w[i]);
	BOOST_CHECK_MESSAGE( are_equal, "contents of v and w differ at position " << i );

	BOOST_MESSAGE( "encode_ternary..." );
	coder::ternary::encode(v,z);
	BOOST_MESSAGE( "decode_ternary..." );
	coder::ternary::decode(z,w);

	are_equal  = true;
	i=0;
	BOOST_CHECK_EQUAL( v.size(), w.size() );
	BOOST_CHECK_EQUAL( v.getIntWidth(), w.getIntWidth() );
	for(; i < v.size() && are_equal; ++i)
		are_equal  = (v[i]==w[i]);
	BOOST_CHECK_MESSAGE( are_equal, "contents of v and w differ at position " << i );

	v.setIntWidth(1);
	v.resize( 1 * v.getIntWidth() );
	v[0] = 0;
	// 0 cannot be encoded
	BOOST_CHECK_THROW( coder::fibonacci::encode(v,z), std::logic_error  );
	BOOST_CHECK_THROW( coder::elias_delta::encode(v,z), std::logic_error  );
	BOOST_CHECK_THROW( coder::ternary::encode(v,z), std::logic_error  );

	int_vector<> ps(10, 0, 32), ps2;
	for(size_t i=0;i<ps.size();++i)
		ps[i] = i+1+(rand()%18);
	ps[ps.size()-1] = ps[ps.size()-1]+1000000000;
	coder::ternary::encode(ps,z);
	coder::ternary::decode(z,ps2);
	BOOST_CHECK_EQUAL(  ps.size(), ps2.size() );
	BOOST_CHECK_EQUAL(  ps.getIntWidth(), ps2.getIntWidth() );
	are_equal = true;
	for(; i < ps.size() && are_equal; ++i)
		are_equal  = (ps[i]==ps2[i]);
	BOOST_CHECK_MESSAGE( are_equal, "contents of ps and ps2 differ at position " << i );


	for(size_t i=0, prefix_sum=0;i<ps.size();++i){
		prefix_sum += ps[i];
		uint64_t temp;
		BOOST_CHECK( (temp=coder::ternary::decode_prefix_sum(z.data(), 0, i+1)) == prefix_sum );
		if( temp != prefix_sum ){
			BOOST_MESSAGE("prefix_sum = "<<prefix_sum<<" calculated = "<<temp);
		}
	}


}


BOOST_AUTO_TEST_CASE( rle_coding_test ){
	const int size=10;
	uint64_t *data = new uint64_t[size];
	for(int i=1; i<=10;++i)
		data[i-1] = i;
	BOOST_CHECK( coder::run_length<coder::fibonacci>::encoding_length(data, 0, 64) == coder::fibonacci::encoding_length(1) + coder::fibonacci::encoding_length(63) );
	BOOST_CHECK( coder::run_length<coder::fibonacci>::encoding_length(data, 0, 65) == coder::fibonacci::encoding_length(1) + coder::fibonacci::encoding_length(64) );
	BOOST_CHECK( coder::run_length<coder::fibonacci>::encoding_length(data, 0, 66) == (uint64_t)2*coder::fibonacci::encoding_length(1) + coder::fibonacci::encoding_length(64) );
	BOOST_CHECK( coder::run_length<coder::fibonacci>::encoding_length(data, 0, 128) == (uint64_t)2*coder::fibonacci::encoding_length(1) + 2*coder::fibonacci::encoding_length(63) );
	delete data;
}

BOOST_AUTO_TEST_SUITE_END()
