#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MAIN 
#include <boost/test/unit_test.hpp>
#include <sdsl/vectors.hpp>
#include <algorithm> // for stl conform test 
#include <fstream>
#include <vector>
#if BOOST_WORKAROUND( __GNUC__, < 3 )
typedef boost::test_tools::output_test_stream onullstream_type;
#else
#include <boost/test/utils/nullstream.hpp>
typedef boost::onullstream onullstream_type;
#endif

using namespace boost::unit_test;
using namespace sdsl;

BOOST_AUTO_TEST_SUITE( int_vector_test_suite )

BOOST_AUTO_TEST_CASE( constructor_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE( "Running tests for the constructor of int_vector..." );
//	std::ofstream ofs("testcase.log");
//	unit_test_log.set_stream( ofs );
	BOOST_MESSAGE(" int_vector constructor test");
	unit_test_log.set_stream( std::cerr );
	bit_vector* S = new bit_vector(100);

	BOOST_MESSAGE("S[0]="<< S[0] );
	delete S;
	//BOOST_MESSAGE("s[0]=" << s[0]);

	// Test default constructor
	int_vector<> v;
	BOOST_CHECK_EQUAL( v.bit_size(), (size_t)0 );
	BOOST_CHECK( v.bit_size() <= v.capacity() );
	BOOST_CHECK_EQUAL( v.getIntWidth(), (uint8_t)64 );

	// Test constructor with parameter size specified
	size_t size = 1000000;
	int_vector<> w(size);
	util::set_random_bits(w);
	size_t ones 	= util::get_one_bits( w );
	if( ones >= w.bit_size()/2 ){
		BOOST_WARN_MESSAGE(ones-w.bit_size()/2 < w.bit_size()/200 + 1000 ,"Maybe setRandomBits does not work well "<<ones<<" "<<size);
	}
	else{
		BOOST_WARN_MESSAGE(w.bit_size()/2-ones < w.bit_size()/200 + 1000 ,"Maybe setRandomBits does not work well "<<ones<<" "<<size);
	}
	BOOST_CHECK_EQUAL( w.bit_size(), size*64 );
	BOOST_CHECK( w.bit_size() <= w.capacity() );
	BOOST_CHECK_EQUAL( w.getIntWidth(), (uint8_t)64);

	onullstream_type devnullout;
	size_t wssize = w.serialize(devnullout);
	BOOST_MESSAGE( "wssize = " << wssize );


	// Test constructor with parameter size and intWidth specified
	uint8_t intWidth = 26;
	int_vector<> x(size, 0, intWidth);
	BOOST_CHECK_EQUAL( x.bit_size(), size*intWidth );
	BOOST_CHECK( x.bit_size() <= x.capacity() );
	BOOST_CHECK_EQUAL( x.getIntWidth(), intWidth);

	// Test copy constructor
	int_vector<> z(w);
	BOOST_CHECK_EQUAL( z.bit_size(), w.bit_size() );
	BOOST_CHECK_EQUAL( z.getIntWidth(), w.getIntWidth() );
	bool are_equal = true;
	size_t i = 0;
	for(; i < z.size() && are_equal; ++i)
		are_equal  = (z[i]==w[i]);
	BOOST_CHECK_MESSAGE( are_equal, "contents of z and w differe at position " << i );

	int_vector<> iv(10,45);
	for(int_vector<>::size_type i=0; i<iv.size(); ++i)
		BOOST_CHECK( 45 == iv[i] );

	int_vector<18> iv2(100, 531);
	for(int_vector<>::size_type i=0; i<iv2.size(); ++i)
		BOOST_CHECK( 531 == iv2[i] );
}

BOOST_AUTO_TEST_CASE( serialize_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE( "Running tests for serialize and load of int_vector..." );
	const char *file_name = "bit_vector_serialize_test.data";
	int_vector<> v(10000);
	util::set_random_bits(v);
	// serialize v and store it to a file
	BOOST_CHECK ( util::store_to_file(v, file_name) );
	int_vector<> w;
	// load vector from file
	BOOST_CHECK( util::load_from_file(w, file_name) );
	BOOST_CHECK(v==w);
	//remove file
	std::remove(file_name);
}

BOOST_AUTO_TEST_CASE( operator_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE( "Running tests for operators of int_vector..." );
	size_t s = 100000;
	int_vector<> v;
	int_vector<> w(s,0, 10);
	util::set_random_bits(w);
	// Test operator =
	v = w;
	BOOST_CHECK_EQUAL( v.bit_size(), w.bit_size() );
	BOOST_CHECK_EQUAL( v.getIntWidth(), w.getIntWidth() );
	bool are_equal = true;
	size_t i = 0;
	for(; i < v.size() && are_equal; ++i)
		are_equal  = (v[i]==w[i]);
	BOOST_CHECK_MESSAGE( are_equal, "contents of v and w differe at bit position " << i );
	// Test operator ==
	BOOST_CHECK( v == w );
	// Test opertor !=
	BOOST_CHECK( !(v != w) );

	const int_vector<> c(100);
	int_vector<>::const_iterator it = c.begin();
}


BOOST_AUTO_TEST_CASE( data_access_test )
{
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE( "Running tests for data access of int_vector..." );
	size_t s = 100000;
	int_vector<> v(s,0,1);
	util::set_one_bits(v);
	BOOST_CHECK_EQUAL( (uint64_t)v[0], (uint64_t)1 );
	v[0] = 0;
	BOOST_CHECK_EQUAL( (uint64_t)v[0], (uint64_t)0 );
	v[1]=v[2]=v[3]=0;
	BOOST_CHECK_EQUAL( v[1]+v[2]+v[3], (uint64_t)0 );
	v[1]=v[2]=v[3]=1;
	BOOST_CHECK_EQUAL( v[1]+v[2]+v[3], (uint64_t)3 );
}

BOOST_AUTO_TEST_CASE( stl_conform_test ){
	unit_test_log.set_threshold_level( log_messages );
	BOOST_TEST_MESSAGE( "Running tests for stl conformance of int_vector..." );

	size_t s = 15;
	int_vector<> v(s, 0, bit_magic::l1BP(s)+1 );
	int_vector<> w,x;
	BOOST_MESSAGE("width "<< (int)v.getIntWidth() );
	BOOST_MESSAGE("size "<< (int)v.size() );
	size_t i=0;
	for(int_vector<>::iterator it = v.begin(); it != v.end(); ++it){
		*it = i++; 
	}	
	w = v;
	reverse(w.begin(), w.end());
	x = w;
	reverse(x.begin(), x.end());
	BOOST_CHECK( v == x );

	sort(w.begin(), w.end());
	BOOST_CHECK( v == w );

	random_shuffle(w.begin(), w.end());
	
//	const SDSBitVector *y = &w;
//	size_t cnt=0;
//	SDSBitVector::const_iterator begin(y,0), end(y, y->bit_size());
//		BOOST_TEST_MESSAGE(" " << "diff " << " "<< ( (uint64_t*)end.m_word - (uint64_t*)begin.m_word));
//	for(SDSBitVector::const_iterator it=y->begin(); it != y->end(); ++it){
//		BOOST_TEST_MESSAGE(" " << *it );
//		BOOST_TEST_MESSAGE(" " << *it << " "<<y->end().m_offset<<" "<< (y->end().m_word - y->data()));
//		BOOST_TEST_MESSAGE(" " << "diff " << " "<<y->end().m_word - y->begin().m_word);
//	}
}

BOOST_AUTO_TEST_CASE( util_test ){
	bit_vector b(70);
	typedef bit_vector::size_type size_type;
	for(size_type i=0; i<70; ++i) b[i] = 1;
	BOOST_REQUIRE( util::get_one_bits(b) == 70 );
//	BOOST_MESSAGE( "getOneZeroBits = " << util::getOneZeroBits(b) );
	BOOST_REQUIRE( util::get_onezero_bits(b) == 0 );
//	BOOST_MESSAGE( "getZeroOneBits = " << util::getZeroOneBits(b) );
	BOOST_REQUIRE( util::get_zeroone_bits(b) == 0 );

	b[64] = 0;
	BOOST_REQUIRE( util::get_onezero_bits(b) == 1 );
	BOOST_REQUIRE( util::get_zeroone_bits(b) == 1 );
	b[69] = 0;
	BOOST_REQUIRE( util::get_onezero_bits(b) == 2 );
	BOOST_REQUIRE( util::get_zeroone_bits(b) == 1 );
}

BOOST_AUTO_TEST_SUITE_END()
